"""Types for styx generated wrappers."""

import pathlib
import shlex
import typing

InputPathType: typing.TypeAlias = pathlib.Path | str
"""Input host file type."""
OutputPathType: typing.TypeAlias = pathlib.Path
"""Output host file type."""


class Execution(typing.Protocol):
    """Execution object used to execute commands.

    Created by `Runner.start_execution()`.
    """

    def input_file(self, host_file: InputPathType) -> str:
        """Resolve host input files.

        Returns a local filepath.
        Called (potentially multiple times) after
        `Runner.start_execution()` and before `Runner.run()`.
        """
        ...

    def run(self, cargs: list[str]) -> None:
        """Run the command.

        Called after all `Execution.input_file()`
        and `Execution.output_file()` calls.
        """
        ...

    def output_file(self, local_file: str, optional: bool = False) -> OutputPathType:
        """Resolve local output files.

        Returns a host filepath.
        Called (potentially multiple times) after all
        `Runner.input_file()` calls.
        """
        ...


class Metadata(typing.NamedTuple):
    """Static tool metadata.

    This is structured static metadata that is known at compile time.
    Runners can use this to set up execution environments.
    """

    id: str
    """Unique identifier of the tool."""
    name: str
    """Name of the tool."""
    package: str | None = None
    """Name of the package that provides the tool."""
    citations: list[str] | None = None
    """List of references to cite when using the tool."""
    container_image_type: str | None = None
    """Type of container image. Example: docker, singularity."""
    container_image_tag: str | None = None
    """Name of an image where the tool is installed and configured.
    Example: bids/mriqc.
    """
    container_image_index: str | None = None
    """Optional index where the image is available, if not the standard location.
    Example: docker.io
    """
    container_image_opts: str | None = None
    """Container-level arguments for the application.
    Example: --privileged
    """


class Runner(typing.Protocol):
    """Runner object used to execute commands.

    Possible examples would be `LocalRunner`,
    `DockerRunner`, `DebugRunner`, ...
    Used as a factory for `Execution` objects.
    """

    def start_execution(self, metadata: Metadata) -> Execution:
        """Start an execution.

        Called before any `Execution.input_file()` calls.
        """
        ...


class StyxRuntimeError(Exception):
    """Styx runtime error.

    Raised when a command reports a non-zero return code.
    """

    def __init__(
        self,
        return_code: int | None = None,
        command_args: list[str] | None = None,
        message_extra: str | None = None,
    ) -> None:
        """Initialize the error."""
        self.return_code = return_code
        self.command_args = command_args

        if return_code is not None:
            message = f"Command failed with return code {return_code}."
        else:
            message = "Command failed."

        if command_args is not None:
            message += f"\n- Command args: {shlex.join(command_args)}"

        if message_extra is not None:
            message += f"\n{message_extra}"

        super().__init__(message)



"""Global state for the Styx library."""

from .local_runner import LocalRunner
from .types import Runner

_STYX_GLOBAL_RUNNER: Runner | None = None


def get_global_runner() -> Runner:
    """Get the default runner."""
    global _STYX_GLOBAL_RUNNER
    if _STYX_GLOBAL_RUNNER is None:
        _STYX_GLOBAL_RUNNER = LocalRunner()
    return _STYX_GLOBAL_RUNNER


def set_global_runner(runner: Runner) -> None:
    """Set the default runner."""
    global _STYX_GLOBAL_RUNNER
    _STYX_GLOBAL_RUNNER = runner



"""Simple local runner implementation."""

import logging
import os
import pathlib
import shlex
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import partial
from subprocess import PIPE, Popen

from .types import (
    Execution,
    InputPathType,
    Metadata,
    OutputPathType,
    Runner,
    StyxRuntimeError,
)


class _LocalExecution(Execution):
    """Local execution object."""

    def __init__(
        self,
        logger: logging.Logger,
        output_dir: pathlib.Path,
        metadata: Metadata,
    ) -> None:
        """Initialize the execution."""
        self.logger: logging.Logger = logger
        self.output_dir: pathlib.Path = output_dir
        self.metadata: Metadata = metadata

        while self.output_dir.exists():
            self.logger.warning(
                f"Output directory {self.output_dir} already exists. Trying another."
            )
            self.output_dir = self.output_dir.with_name(f"{self.output_dir.name}_1")
        self.output_dir.mkdir(parents=True, exist_ok=True)

    def input_file(self, host_file: InputPathType) -> str:
        """Resolve host input files."""
        return str(pathlib.Path(host_file).absolute())

    def output_file(self, local_file: str, optional: bool = False) -> OutputPathType:
        """Resolve local output files."""
        return self.output_dir / local_file

    def run(self, cargs: list[str]) -> None:
        """Run the command."""
        self.logger.debug(f"Running command: {shlex.join(cargs)}")

        def _stdout_handler(line: str) -> None:
            self.logger.info(line)

        def _stderr_handler(line: str) -> None:
            self.logger.error(line)

        time_start = datetime.now()
        with Popen(
            cargs, text=True, stdout=PIPE, stderr=PIPE, cwd=self.output_dir
        ) as process:
            with ThreadPoolExecutor(2) as pool:  # two threads to handle the streams
                exhaust = partial(pool.submit, partial(deque, maxlen=0))
                exhaust(_stdout_handler(line[:-1]) for line in process.stdout)  # type: ignore
                exhaust(_stderr_handler(line[:-1]) for line in process.stderr)  # type: ignore
        return_code = process.poll()
        time_end = datetime.now()
        self.logger.info(f"Executed {self.metadata.name} in {time_end - time_start}")
        if return_code:
            raise StyxRuntimeError(return_code, cargs)


class LocalRunner(Runner):
    """Local runner implementation."""

    logger_name = "styx_local_runner"

    def __init__(self, data_dir: InputPathType | None = None) -> None:
        """Initialize the runner."""
        self.data_dir = pathlib.Path(data_dir or "styx_tmp")
        self.uid = os.urandom(8).hex()
        self.execution_counter = 0

        # Configure logger
        self.logger = logging.getLogger(self.logger_name)
        if not self.logger.hasHandlers():
            self.logger.setLevel(logging.DEBUG)
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter("[%(levelname).1s] %(message)s")
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)

    def start_execution(self, metadata: Metadata) -> Execution:
        """Start a new execution."""
        output_dir = (
            self.data_dir / f"{self.uid}_{self.execution_counter}_{metadata.name}"
        )
        self.execution_counter += 1
        return _LocalExecution(
            logger=self.logger,
            output_dir=output_dir,
            metadata=metadata,
        )



""".. include:: ../../README.md"""  # noqa: D415

from .global_state import (
    get_global_runner,
    set_global_runner,
)
from .local_runner import (
    LocalRunner,
)
from .types import (
    Execution,
    InputPathType,
    Metadata,
    OutputPathType,
    Runner,
    StyxRuntimeError,
)

__all__ = [
    "Execution",
    "InputPathType",
    "Metadata",
    "OutputPathType",
    "Runner",
    "LocalRunner",
    "get_global_runner",
    "set_global_runner",
    "StyxRuntimeError",
]
